class Event:
    """Event Singleton Class

    How to use example:
    from .event import Event
    Event().profile_data_updated(400, 312)
    """

    singleton = None
    callbacks = dict()

    def __new__(cls, *args, **kwargs):
        if not cls.singleton:
            cls.singleton = object.__new__(Event)
        return cls.singleton

    def __init__(self):
        pass

    def fire_callbacks(self, function_name, *args, **kwargs):
        if function_name not in self.callbacks:
            return
        for callback in self.callbacks[function_name]:
            callback(*args, **kwargs)

    def add_callback(self, function_name, callback):
        if function_name not in self.callbacks:
            self.callbacks[function_name] = []

        self.callbacks[function_name].append(callback)

    # place custom events below
    def profile_data_updated(self, username, followers_count, following_count):
        self.fire_callbacks(
            self.profile_data_updated.__name__,
            username,
            followers_count,
            following_count,
        )

    def commented(self, username):
        self.fire_callbacks(self.commented.__name__, username)

    def liked(self, username):
        self.fire_callbacks(self.liked.__name__, username)

    def followed(self, username):
        self.fire_callbacks(self.followed.__name__, username)

    def unfollowed(self, username):
        self.fire_callbacks(self.unfollowed.__name__, username)



import requests
import sqlite3
from .settings import Settings
from .database_engine import get_database


def get_server_endpoint(topic):
    if topic == "fashion":
        return Settings.pods_fashion_server_endpoint
    elif topic == "food":
        return Settings.pods_food_server_endpoint
    elif topic == "travel":
        return Settings.pods_travel_server_endpoint
    elif topic == "sports":
        return Settings.pods_sports_server_endpoint
    elif topic == "entertainment":
        return Settings.pods_entertainment_server_endpoint
    else:
        return Settings.pods_server_endpoint


def get_recent_posts_from_pods(topic, logger):
    """ fetches all recent posts shared with pods """
    params = {"topic": topic}
    r = requests.get(get_server_endpoint(topic) + "/getRecentPostsV1", params=params)
    try:
        logger.info("Downloaded postids from Pod {}:".format(topic))
        if r.status_code == 200:
            logger.info(r.json())
            return r.json()
        else:
            logger.error(r.text)
            return []
    except Exception as err:
        logger.error("Could not get postids from pod {} - {}".format(topic, err))
        return []


def group_posts(posts, logger):
    no_comments_post_ids = []
    light_post_ids = []
    normal_post_ids = []
    heavy_post_ids = []
    for postobj in posts:
        try:
            if postobj["mode"] == "no_comments":
                no_comments_post_ids.append(postobj)
            elif postobj["mode"] == "light":
                light_post_ids.append(postobj)
            elif postobj["mode"] == "heavy":
                heavy_post_ids.append(postobj)
            else:
                normal_post_ids.append(postobj)
        except Exception as err:
            logger.error(
                "Failed with Error {}, please upgrade your instapy".format(err)
            )
            normal_post_ids.append(postobj)
    return no_comments_post_ids, light_post_ids, normal_post_ids, heavy_post_ids


def share_my_post_with_pods(postid, topic, engagement_mode, logger):
    """ share_my_post_with_pod """
    params = {"postid": postid, "topic": topic, "mode": engagement_mode}
    r = requests.get(get_server_endpoint(topic) + "/publishMyLatestPost", params=params)
    try:
        logger.info("Publishing to Pods {}".format(postid))
        if r.status_code == 200:
            logger.info(r.text)
            return True
        else:
            logger.error(r.text)
            return False
    except Exception as err:
        logger.error(err)
        return False


def share_with_pods_restriction(operation, postid, limit, logger):
    """ Keep track of already shared posts """
    try:
        # get a DB and start a connection
        db, id = get_database()
        conn = sqlite3.connect(db)

        with conn:
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()

            cur.execute(
                "SELECT * FROM shareWithPodsRestriction WHERE profile_id=:id_var "
                "AND postid=:name_var",
                {"id_var": id, "name_var": postid},
            )
            data = cur.fetchone()
            share_data = dict(data) if data else None

            if operation == "write":
                if share_data is None:
                    # write a new record
                    cur.execute(
                        "INSERT INTO shareWithPodsRestriction (profile_id, "
                        "postid, times) VALUES (?, ?, ?)",
                        (id, postid, 1),
                    )
                else:
                    # update the existing record
                    share_data["times"] += 1
                    sql = (
                        "UPDATE shareWithPodsRestriction set times = ? WHERE "
                        "profile_id=? AND postid = ?"
                    )
                    cur.execute(sql, (share_data["times"], id, postid))

                # commit the latest changes
                conn.commit()

            elif operation == "read":
                if share_data is None:
                    return False

                elif share_data["times"] < limit:
                    return False

                else:
                    exceed_msg = "" if share_data["times"] == limit else "more than "
                    logger.info(
                        "---> {} has already been shared with pods {}{} times".format(
                            postid, exceed_msg, str(limit)
                        )
                    )
                    return True

    except Exception as exc:
        logger.error(
            "Dap! Error occurred with share Restriction:\n\t{}".format(
                str(exc).encode("utf-8")
            )
        )

    finally:
        if conn:
            # close the open connection
            conn.close()


def comment_restriction(operation, postid, limit, logger):
    """ Keep track of already shared posts """
    try:
        # get a DB and start a connection
        db, id = get_database()
        conn = sqlite3.connect(db)

        with conn:
            conn.row_factory = sqlite3.Row
            cur = conn.cursor()

            cur.execute(
                "SELECT * FROM commentRestriction WHERE profile_id=:id_var "
                "AND postid=:name_var",
                {"id_var": id, "name_var": postid},
            )
            data = cur.fetchone()
            share_data = dict(data) if data else None

            if operation == "write":
                if share_data is None:
                    # write a new record
                    cur.execute(
                        "INSERT INTO commentRestriction (profile_id, "
                        "postid, times) VALUES (?, ?, ?)",
                        (id, postid, 1),
                    )
                else:
                    # update the existing record
                    share_data["times"] += 1
                    sql = (
                        "UPDATE commentRestriction set times = ? WHERE "
                        "profile_id=? AND postid = ?"
                    )
                    cur.execute(sql, (share_data["times"], id, postid))

                # commit the latest changes
                conn.commit()

            elif operation == "read":
                if share_data is None:
                    return False

                elif share_data["times"] < limit:
                    return False

                else:
                    exceed_msg = "" if share_data["times"] == limit else "more than "
                    logger.info(
                        "---> {} has been commented on {}{} times".format(
                            postid, exceed_msg, str(limit)
                        )
                    )
                    return True

    except Exception as exc:
        logger.error(
            "Dap! Error occurred with comment Restriction:\n\t{}".format(
                str(exc).encode("utf-8")
            )
        )

    finally:
        if conn:
            # close the open connection
            conn.close()



import time
import math
from random import randint
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import WebDriverException
from .util import click_element
from .util import web_address_navigator
from .util import update_activity
from .util import get_action_delay
from .settings import Settings
from .xpath import read_xpath

import requests


def get_story_data(browser, elem, action_type, logger, simulate=False):
    """
    get the JSON data from the graphql URL
    output the amount of segments we can watch
    """

    # if things change in the future, modify here:
    query_hash = "cda12de4f7fd3719c0569ce03589f4c4"
    elem_id = ""

    if action_type == "user":
        try:
            reel_id = browser.execute_script(
                "return window.__additionalData[Object.keys(window.__additionalData)[0]].data.graphql.user.id"
            )
            # correct formating for elem_id
            elem_id = '"' + reel_id + '"'
            # and elem needs to be nothing
            elem = ""
        except WebDriverException:
            try:
                reel_id = browser.execute_script(
                    "return window._sharedData."
                    "entry_data.ProfilePage[0]."
                    "graphql.user.id"
                )
                # correct formating for elem_id
                elem_id = '"' + reel_id + '"'
                # and elem needs to be nothing
                elem = ""
            except WebDriverException:
                logger.error(
                    "---> Sorry, this page isn't available!\t~either "
                    + "link is broken or page is removed\n"
                )
                return {"status": "not_ok", "reels_cnt": 0}
    else:
        reel_id = "tag:{}".format(elem)

    graphql_query_url = (
        "https://www.instagram.com/graphql/query/?query_hash={}"
        '&variables={{"reel_ids":[{}],"tag_names":["{}"],"location_ids":[],'
        '"highlight_reel_ids":[],"precomposed_overlay":false,"show_story_viewer_list":true,'
        '"story_viewer_fetch_count":50,"story_viewer_cursor":"",'
        '"stories_video_dash_manifest":false}}'.format(query_hash, elem_id, elem)
    )

    cookies = browser.get_cookies()
    session = requests.Session()
    csrftoken = ""

    # prepare the cookies for the requests session
    for cookie in cookies:
        all_args = {
            "name": cookie["name"],
            "value": cookie["value"],
            "domain": cookie["domain"],
            "secure": cookie["secure"],
            "rest": {"HttpOnly": cookie["httpOnly"]},
            "path": cookie["path"],
        }
        if cookie["name"] == "csrftoken":
            csrftoken = cookie["value"]
        if not (cookie["name"] == "urlgen") and not (cookie["name"] == "rur"):
            all_args["expires"] = cookie["expiry"]

        session.cookies.set(**all_args)

    headers = {"User-Agent": Settings.user_agent, "X-Requested-With": "XMLHttpRequest"}

    data = session.get(graphql_query_url, headers=headers)
    response = data.json()
    update_activity(browser, state=None)

    reels_cnt = 0
    if response["status"] == "ok":
        # we got a correct response from the server
        # check how many reels we got
        media_cnt = len(response["data"]["reels_media"])

        if media_cnt == 0:
            # then nothing to watch, we received no stories
            return {"status": "ok", "reels_cnt": 0}
        else:
            # we got content
            # check if there is something new to watch otherwise we just return 0
            seen = 0
            if (action_type != "tag") and (
                response["data"]["reels_media"][0]["seen"] is not None
            ):
                seen = response["data"]["reels_media"][0]["seen"]
            index = 1
            if simulate is True:
                for item in response["data"]["reels_media"][0]["items"]:
                    if item["taken_at_timestamp"] <= seen:
                        continue
                    else:
                        headers = {
                            "User-Agent": Settings.user_agent,
                            "X-CSRFToken": csrftoken,
                            "X-Requested-With": "XMLHttpRequest",
                            "Content-Type": "application/x-www-form-urlencoded",
                        }
                        response = session.post(
                            "https://www.instagram.com/stories/reel/seen",
                            data={
                                "reelMediaId": item["id"],
                                "reelMediaOwnerId": item["owner"]["id"],
                                "reelId": reel_id,
                                "reelMediaTakenAt": item["taken_at_timestamp"],
                                "viewSeenAt": math.floor(time.time()),
                            },
                            headers=headers,
                        )
                        logger.info("  --> simulated watch reel # {}".format(index))
                        update_activity()
                        index += 1
                        time.sleep(randint(3, 6))
                        reels_cnt += 1
            else:
                story_elem = browser.find_element_by_xpath(
                    read_xpath(
                        watch_story.__name__ + "_for_{}".format(action_type),
                        "explore_stories",
                    )
                )

                click_element(browser, story_elem)

                logger.info("Watching stories...")
                for item in response["data"]["reels_media"][0]["items"]:
                    if item["taken_at_timestamp"] <= seen:
                        continue
                    else:
                        time.sleep(2)
                        if index == 1:
                            try:
                                next_elem = browser.find_element_by_xpath(
                                    read_xpath(watch_story.__name__, "next_first")
                                )
                            except NoSuchElementException:
                                continue
                        else:
                            try:
                                next_elem = browser.find_element_by_xpath(
                                    read_xpath(watch_story.__name__, "next")
                                )
                            except NoSuchElementException:
                                continue
                        click_element(browser, next_elem)
                        reels_cnt += 1
                        index += 1

            return {"status": "ok", "reels_cnt": reels_cnt}
    else:
        return {"status": "not_ok", "reels_cnt": 0}


def watch_story(browser, elem, logger, action_type, simulate=False):
    """
        Load Stories, and watch it until there is no more stores
        to watch for the related element
    """

    # make sure we work with a lower case elem
    elem = elem.lower()

    if action_type == "tag":
        story_link = "https://www.instagram.com/explore/tags/{}".format(elem)

    else:
        story_link = "https://www.instagram.com/{}".format(elem)

    web_address_navigator(browser, story_link)
    # wait for the page to load
    time.sleep(randint(2, 6))
    # order is important here otherwise we are not on the page of the story we want to watch
    story_data = get_story_data(browser, elem, action_type, logger, simulate)

    if story_data["status"] == "not ok":
        raise NoSuchElementException

    if story_data["reels_cnt"] == 0:
        # nothing to watch, there is no stories
        logger.info(
            "no stories to watch (either there is none) or we have already watched everything"
        )
        return 0

    logger.info(
        "watched {} reels from {}: {}".format(
            story_data["reels_cnt"], action_type, elem.encode("utf-8")
        )
    )

    # get the post-story delay time to sleep
    naply = get_action_delay("story")
    time.sleep(naply)

    return story_data["reels_cnt"]



"""Module only used for the login part of the script"""
# import built-in & third-party modules
import pickle
import socket
import os
import json
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys

# import InstaPy modules
from .time_util import sleep
from .util import update_activity
from .util import web_address_navigator
from .util import explicit_wait
from .util import click_element
from .util import check_authorization
from .util import reload_webpage

# import exceptions
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import WebDriverException
from selenium.common.exceptions import MoveTargetOutOfBoundsException

from .xpath import read_xpath


def bypass_suspicious_login(
    browser, logger, logfolder, bypass_security_challenge_using
):
    """ Bypass suspicious loggin attempt verification. """

    # close sign up Instagram modal if available
    dismiss_get_app_offer(browser, logger)
    dismiss_notification_offer(browser, logger)
    dismiss_this_was_me(browser)

    option = None
    if bypass_security_challenge_using == "sms":
        try:
            option = browser.find_element_by_xpath(
                read_xpath(bypass_suspicious_login.__name__, "bypass_with_sms_option")
            )
        except NoSuchElementException:
            logger.warn(
                "Unable to choose ({}) option to bypass the challenge".format(
                    bypass_security_challenge_using.upper()
                )
            )

    if bypass_security_challenge_using == "email":
        try:
            option = browser.find_element_by_xpath(
                read_xpath(bypass_suspicious_login.__name__, "bypass_with_email_option")
            )
        except NoSuchElementException:
            logger.warn(
                "Unable to choose ({}) option to bypass the challenge".format(
                    bypass_security_challenge_using.upper()
                )
            )

    # click on your option
    (ActionChains(browser).move_to_element(option).click().perform())
    # next button click will miss the DOM reference for this element, so ->
    option_text = option.text

    # click on security code
    send_security_code_button = browser.find_element_by_xpath(
        read_xpath(bypass_suspicious_login.__name__, "send_security_code_button")
    )
    (ActionChains(browser).move_to_element(send_security_code_button).click().perform())

    # update server calls
    update_activity(browser, state=None)

    print("Instagram detected an unusual login attempt")
    print('Check Instagram App for "Suspicious Login attempt" prompt')
    print("A security code was sent to your {}".format(option_text))

    security_code = None
    try:
        path = "{}state.json".format(logfolder)
        data = {}
        # check if file exists and has content
        if os.path.isfile(path) and os.path.getsize(path) > 0:
            # load JSON file
            with open(path, "r") as json_file:
                data = json.load(json_file)

        # update connection state
        security_code = data["challenge"]["security_code"]
    except Exception:
        logger.info("Security Code not present in {}state.json file".format(logfolder))

    if security_code is None:
        security_code = input("Type the security code here: ")

    security_code_field = browser.find_element_by_xpath(
        read_xpath(bypass_suspicious_login.__name__, "security_code_field")
    )

    (
        ActionChains(browser)
        .move_to_element(security_code_field)
        .click()
        .send_keys(security_code)
        .perform()
    )

    # update server calls for both 'click' and 'send_keys' actions
    for _ in range(2):
        update_activity(browser, state=None)

    submit_security_code_button = browser.find_element_by_xpath(
        read_xpath(bypass_suspicious_login.__name__, "submit_security_code_button")
    )

    (
        ActionChains(browser)
        .move_to_element(submit_security_code_button)
        .click()
        .perform()
    )

    # update server calls
    update_activity(browser, state=None)

    try:
        sleep(3)
        # locate wrong security code message
        wrong_login = browser.find_element_by_xpath(
            read_xpath(bypass_suspicious_login.__name__, "wrong_login")
        )

        if wrong_login is not None:
            wrong_login_msg = (
                "Wrong security code! Please check the code Instagram"
                "sent you and try again."
            )
            update_activity(
                browser,
                action=None,
                state=wrong_login_msg,
                logfolder=logfolder,
                logger=logger,
            )
            print(wrong_login_msg)

    except NoSuchElementException:
        # correct security code
        pass


def check_browser(browser, logfolder, logger, proxy_address):
    # set initial state to offline
    update_activity(
        browser,
        action=None,
        state="trying to connect",
        logfolder=logfolder,
        logger=logger,
    )

    # check connection status
    try:
        logger.info("-- Connection Checklist [1/2] (Internet Connection Status)")
        browser.get("view-source:https://ip4.seeip.org/geoip")
        pre = browser.find_element_by_tag_name("pre").text
        current_ip_info = json.loads(pre)
        if (
            proxy_address is not None
            and socket.gethostbyname(proxy_address) != current_ip_info["ip"]
        ):
            logger.warn("- Proxy is set, but it's not working properly")
            logger.warn(
                '- Expected Proxy IP is "{}", and the current IP is "{}"'.format(
                    proxy_address, current_ip_info["ip"]
                )
            )
            logger.warn("- Try again or disable the Proxy Address on your setup")
            logger.warn("- Aborting connection...")
            return False
        else:
            logger.info("- Internet Connection Status: ok")
            logger.info(
                '- Current IP is "{}" and it\'s from "{}/{}"'.format(
                    current_ip_info["ip"],
                    current_ip_info["country"],
                    current_ip_info["country_code"],
                )
            )
            update_activity(
                browser,
                action=None,
                state="Internet connection is ok",
                logfolder=logfolder,
                logger=logger,
            )
    except Exception:
        logger.warn("- Internet Connection Status: error")
        update_activity(
            browser,
            action=None,
            state="There is an issue with the internet connection",
            logfolder=logfolder,
            logger=logger,
        )
        return False

    # check if hide-selenium extension is running
    logger.info("-- Connection Checklist [2/2] (Hide Selenium Extension)")
    webdriver = browser.execute_script("return window.navigator.webdriver")
    logger.info("- window.navigator.webdriver response: {}".format(webdriver))
    if webdriver:
        logger.warn("- Hide Selenium Extension: error")
    else:
        logger.info("- Hide Selenium Extension: ok")

    # everything is ok, then continue(True)
    return True


def login_user(
    browser,
    username,
    password,
    logger,
    logfolder,
    proxy_address,
    security_code_to_phone,
    want_check_browser,
):
    """Logins the user with the given username and password"""
    assert username, "Username not provided"
    assert password, "Password not provided"

    # Hotfix - this check crashes more often than not -- plus in not necessary, I can verify my own connection
    if want_check_browser:
        if not check_browser(browser, logfolder, logger, proxy_address):
            return False

    ig_homepage = "https://www.instagram.com"
    web_address_navigator(browser, ig_homepage)
    cookie_loaded = False

    # try to load cookie from username
    try:
        for cookie in pickle.load(
            open("{0}{1}_cookie.pkl".format(logfolder, username), "rb")
        ):
            browser.add_cookie(cookie)
            cookie_loaded = True
    except (WebDriverException, OSError, IOError):
        print("Cookie file not found, creating cookie...")

    # force refresh after cookie load or check_authorization() will FAIL
    reload_webpage(browser)

    # cookie has been LOADED, so the user SHOULD be logged in
    # check if the user IS logged in
    login_state = check_authorization(
        browser, username, "activity counts", logger, False
    )
    if login_state is True:
        dismiss_notification_offer(browser, logger)
        return True

    # if user is still not logged in, then there is an issue with the cookie
    # so go create a new cookie..
    if cookie_loaded:
        print(
            "Issue with cookie for user {}. Creating " "new cookie...".format(username)
        )

    # Check if the first div is 'Create an Account' or 'Log In'
    try:
        login_elem = browser.find_element_by_xpath(
            read_xpath(login_user.__name__, "login_elem")
        )
    except NoSuchElementException:
        print("Login A/B test detected! Trying another string...")
        try:
            login_elem = browser.find_element_by_xpath(
                read_xpath(login_user.__name__, "login_elem_no_such_exception")
            )
        except NoSuchElementException:
            print("Could not pass the login A/B test. Trying last string...")
            try:
                login_elem = browser.find_element_by_xpath(
                    read_xpath(login_user.__name__, "login_elem_no_such_exception_2")
                )
            except NoSuchElementException:
                return False

    if login_elem is not None:
        try:
            (ActionChains(browser).move_to_element(login_elem).click().perform())
        except MoveTargetOutOfBoundsException:
            login_elem.click()

        # update server calls
        update_activity(browser, state=None)

    # Enter username and password and logs the user in
    # Sometimes the element name isn't 'Username' and 'Password'
    # (valid for placeholder too)

    # wait until it navigates to the login page
    login_page_title = "Login"
    explicit_wait(browser, "TC", login_page_title, logger)

    # wait until the 'username' input element is located and visible
    input_username_XP = read_xpath(login_user.__name__, "input_username_XP")
    explicit_wait(browser, "VOEL", [input_username_XP, "XPath"], logger)

    input_username = browser.find_element_by_xpath(input_username_XP)

    (
        ActionChains(browser)
        .move_to_element(input_username)
        .click()
        .send_keys(username)
        .perform()
    )

    # update server calls for both 'click' and 'send_keys' actions
    for _ in range(2):
        update_activity(browser, state=None)

    sleep(1)

    #  password
    input_password = browser.find_elements_by_xpath(
        read_xpath(login_user.__name__, "input_password")
    )

    if not isinstance(password, str):
        password = str(password)

    (
        ActionChains(browser)
        .move_to_element(input_password[0])
        .click()
        .send_keys(password)
        .perform()
    )

    sleep(1)

    (
        ActionChains(browser)
        .move_to_element(input_password[0])
        .click()
        .send_keys(Keys.ENTER)
        .perform()
    )

    # update server calls for both 'click' and 'send_keys' actions
    for _ in range(4):
        update_activity(browser, state=None)

    dismiss_get_app_offer(browser, logger)
    dismiss_notification_offer(browser, logger)

    # check for login error messages and display it in the logs
    if "instagram.com/challenge" in browser.current_url:
        # check if account is disabled by Instagram,
        # or there is an active challenge to solve
        try:
            account_disabled = browser.find_element_by_xpath(
                read_xpath(login_user.__name__, "account_disabled")
            )
            logger.warn(account_disabled.text)
            update_activity(
                browser,
                action=None,
                state=account_disabled.text,
                logfolder=logfolder,
                logger=logger,
            )
            return False
        except NoSuchElementException:
            pass

        # in case the user doesnt have a phone number linked to the Instagram account
        try:
            browser.find_element_by_xpath(
                read_xpath(login_user.__name__, "add_phone_number")
            )
            challenge_warn_msg = (
                "Instagram initiated a challenge before allow your account to login. "
                "At the moment there isn't a phone number linked to your Instagram "
                "account. Please, add a phone number to your account, and try again."
            )
            logger.warn(challenge_warn_msg)
            update_activity(
                browser,
                action=None,
                state=challenge_warn_msg,
                logfolder=logfolder,
                logger=logger,
            )
            return False
        except NoSuchElementException:
            pass

        # try to initiate security code challenge
        try:
            browser.find_element_by_xpath(
                read_xpath(login_user.__name__, "suspicious_login_attempt")
            )
            update_activity(
                browser,
                action=None,
                state="Trying to solve suspicious attempt login",
                logfolder=logfolder,
                logger=logger,
            )
            bypass_suspicious_login(browser, logger, logfolder, security_code_to_phone)
        except NoSuchElementException:
            pass

    # check for wrong username or password message, and show it to the user
    try:
        error_alert = browser.find_element_by_xpath(
            read_xpath(login_user.__name__, "error_alert")
        )
        logger.warn(error_alert.text)
        update_activity(
            browser,
            action=None,
            state=error_alert.text,
            logfolder=logfolder,
            logger=logger,
        )
        return False
    except NoSuchElementException:
        pass

    if "instagram.com/accounts/onetap" in browser.current_url:
        browser.get("https://instagram.com")

    # wait until page fully load
    explicit_wait(browser, "PFL", [], logger, 5)

    # Check if user is logged-in (If there's two 'nav' elements)
    nav = browser.find_elements_by_xpath(read_xpath(login_user.__name__, "nav"))
    if len(nav) == 2:
        # create cookie for username
        pickle.dump(
            browser.get_cookies(),
            open("{0}{1}_cookie.pkl".format(logfolder, username), "wb"),
        )
        return True
    else:
        return False


def dismiss_get_app_offer(browser, logger):
    """ Dismiss 'Get the Instagram App' page after a fresh login """
    offer_elem = read_xpath(dismiss_get_app_offer.__name__, "offer_elem")
    dismiss_elem = read_xpath(dismiss_get_app_offer.__name__, "dismiss_elem")

    # wait a bit and see if the 'Get App' offer rises up
    offer_loaded = explicit_wait(
        browser, "VOEL", [offer_elem, "XPath"], logger, 5, False
    )

    if offer_loaded:
        dismiss_elem = browser.find_element_by_xpath(dismiss_elem)
        click_element(browser, dismiss_elem)


def dismiss_notification_offer(browser, logger):
    """ Dismiss 'Turn on Notifications' offer on session start """
    offer_elem_loc = read_xpath(dismiss_notification_offer.__name__, "offer_elem_loc")
    dismiss_elem_loc = read_xpath(
        dismiss_notification_offer.__name__, "dismiss_elem_loc"
    )

    # wait a bit and see if the 'Turn on Notifications' offer rises up
    offer_loaded = explicit_wait(
        browser, "VOEL", [offer_elem_loc, "XPath"], logger, 4, False
    )

    if offer_loaded:
        dismiss_elem = browser.find_element_by_xpath(dismiss_elem_loc)
        click_element(browser, dismiss_elem)


def dismiss_this_was_me(browser):
    try:
        # click on "This was me" button if challenge page was called
        this_was_me_button = browser.find_element_by_xpath(
            read_xpath(dismiss_this_was_me.__name__, "this_was_me_button")
        )
        (ActionChains(browser).move_to_element(this_was_me_button).click().perform())
        # update server calls
        update_activity(browser, state=None)
    except NoSuchElementException:
        # no verification needed
        pass



# This Python file uses the following encoding: utf-8

from setuptools import setup
from os import path

# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import and use built-in open()
from io import open as io_open
import re


summary = "Tool for automated Instagram interactions"
project_homepage = "https://github.com/timgrossmann/InstaPy"
here = path.abspath(path.dirname(__file__))


def readall(*args):
    with io_open(path.join(here, *args), encoding="utf-8") as fp:
        return fp.read()


with open("requirements.txt") as f:
    dependencies = f.read().splitlines()

documentation = readall("README.md")
metadata = dict(
    re.findall(r"""__([a-z]+)__ = "([^"]+)""", readall("instapy", "__init__.py"))
)

setup(
    name="instapy",
    version=metadata["version"],
    description=summary,
    long_description=documentation,
    long_description_content_type="text/markdown",
    author="Tim Großmann",
    author_email="contact.timgrossmann@gmail.com",
    maintainer="InstaPy Community at Github",
    license="GPLv3",
    url=project_homepage,
    download_url=(project_homepage + "/archive/master.zip"),
    project_urls={
        "How Tos": (project_homepage + "/tree/master/docs"),
        "Examples": (project_homepage + "/tree/master/quickstart_templates"),
        "Bug Reports": (project_homepage + "/issues"),
        "Funding": "https://www.paypal.me/supportInstaPy",
        "Say Thanks!": "http://saythanks.io/to/uluQulu",
        "Source": (project_homepage + "/tree/master/instapy"),
    },
    packages=["instapy"],
    # include_package_data=True,  # <- packs every data file in the package
    package_data={  # we need only the files below:
        "instapy": [
            "icons/Windows/*.ico",
            "icons/Linux/*.png",
            "icons/Mac/*.icns",
            "firefox_extension/*",
            "plugins/*",
        ]
    },
    keywords=(
        "instapy python instagram automation \
         marketing promotion bot selenium"
    ),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Win32 (MS Windows)",
        "Environment :: MacOS X",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: JavaScript",
        "Programming Language :: SQL",
        "Topic :: Utilities",
        "Topic :: Software Development :: Build Tools",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Natural Language :: English",
    ],
    install_requires=dependencies,
    extras_require={"test": ["tox", "virtualenv", "tox-venv"]},
    python_requires=">=3, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    platforms=["win32", "linux", "linux2", "darwin"],
    zip_safe=False,
    entry_points={"console_scripts": []},
)



""" Module that handles the like features """
from .util import update_activity

from selenium.common.exceptions import NoSuchElementException

LIKE_TAG_CLASS = "coreSpriteHeartOpen"


def get_like_on_feed(browser, amount):
    """
        browser - the selenium browser element
        amount - total amount of likes to perform

        --------------------------------------
        The function takes in the total amount of likes to perform
        and then sends buttons to be liked, if it has run out of like
        buttons it will perform a scroll
    """

    # get like buttons
    likes_performed = 0

    while likes_performed <= amount:
        # get the like buttons
        like_buttons = []

        abort = False
        try:
            like_buttons = browser.find_elements_by_class_name(LIKE_TAG_CLASS)

        except NoSuchElementException:
            print("Unale to find the like buttons, Aborting")
            abort = True

        if abort:
            break

        for button in like_buttons:
            likes_performed += 1

            if not (likes_performed <= amount):
                print("Performed the required number of likes")
                break
            yield button

        print("---> Total Likes uptil now ->", likes_performed)

        browser.execute_script("window.scrollTo(0, document.body.scrollHeight);")
        update_activity(browser, state=None)



"""
Global variables

By design, import no any other local module inside this file.
Vice verse, it'd produce circular dependent imports.
"""

from sys import platform
from os import environ as environmental_variables
from os.path import join as join_path


WORKSPACE = {
    "name": "InstaPy",
    "path": environmental_variables.get("INSTAPY_WORKSPACE"),
}
OS_ENV = (
    "windows" if platform == "win32" else "osx" if platform == "darwin" else "linux"
)


def localize_path(*args):
    """ Join given locations as an OS path """

    if WORKSPACE["path"]:
        path = join_path(WORKSPACE["path"], *args)
        return path

    else:
        return None


class Settings:
    """ Globally accessible settings throughout whole project """

    # locations
    log_location = localize_path("logs")
    database_location = localize_path("db", "instapy.db")

    # set a logger cache outside the InstaPy object to avoid
    # re-instantiation issues
    loggers = {}
    logger = None

    # set current profile credentials for DB operations
    profile = {"id": None, "name": None}

    # hold live Quota Supervisor configuration for global usage
    QS_config = {}

    # store user-defined delay time to sleep after doing actions
    action_delays = {}

    # store configuration of text analytics
    meaningcloud_config = {}
    yandex_config = {}

    # store the parameter for global access
    show_logs = None

    user_agent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
    )

    # state of instantiation of InstaPy
    InstaPy_is_running = False

    # This is where currently the pods server is hosted
    pods_server_endpoint = "https://us-central1-instapy-pods.cloudfunctions.net"
    pods_fashion_server_endpoint = (
        "https://us-central1-instapy-pods-fashion.cloudfunctions.net"
    )
    pods_food_server_endpoint = (
        "https://us-central1-instapy-pods-food.cloudfunctions.net"
    )
    pods_travel_server_endpoint = (
        "https://us-central1-instapy-pods-travel.cloudfunctions.net"
    )
    pods_sports_server_endpoint = (
        "https://us-central1-instapy-pods-sports.cloudfunctions.net"
    )
    pods_entertainment_server_endpoint = (
        "https://us-central1-instapy-pods-entertainment.cloudfunctions.net"
    )


class Storage:
    """ Globally accessible standalone storage """

    # store realtime record activity data
    record_activity = {}



"""Methods to extract the data for the given usernames profile"""
# code created by modification of original code copied from
# https://github.com/timgrossmann/instagram-profilecrawl/blob/master/util
# /extractor.py
import time
from time import sleep
from datetime import datetime, timedelta
import random
import collections
from operator import itemgetter
from selenium.webdriver.common.keys import Keys

from .util import get_number_of_posts
from .util import click_element
from .util import update_activity
from .util import web_address_navigator
from .util import username_url_to_username
from .util import scroll_bottom
from .util import get_users_from_dialog
from .util import progress_tracker
from .util import close_dialog_box

from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
)

from .xpath import read_xpath


def check_exists_by_xpath(browser, xpath):
    try:
        browser.find_element_by_xpath(xpath)
    except NoSuchElementException:
        return False
    return True


def remove_duplicates_preserving_order(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def extract_post_info(browser):
    """Get the information from the current post"""
    web_address_navigator(browser, browser.current_url + "comments/")
    comments = []
    user_commented_list = []
    last_comment_count = 0
    # load all hidden comments
    while check_exists_by_xpath(
        browser, read_xpath(extract_post_info.__name__, "load_more_comments_element")
    ):
        load_more_comments_element = browser.find_element_by_xpath(
            read_xpath(extract_post_info.__name__, "load_more_comments_element")
        )
        click_element(browser, load_more_comments_element)
        sleep(0.5)
        # get comment list
        comment_list = browser.find_element_by_xpath(
            read_xpath(extract_post_info.__name__, "comment_list")
        )
        comments = comment_list.find_elements_by_xpath(
            read_xpath(extract_post_info.__name__, "comments")
        )
        # check instagram comment load bug
        if len(comments) == last_comment_count:
            break
        if (len(comments) - last_comment_count) < 3:
            break
        last_comment_count = len(comments)

    # get all comment list
    comment_list = browser.find_element_by_xpath(
        read_xpath(extract_post_info.__name__, "comment_list")
    )
    comments = comment_list.find_elements_by_xpath(
        read_xpath(extract_post_info.__name__, "comments")
    )

    # get all commenter list
    try:
        for comm in comments:
            user_commented = (
                comm.find_element_by_tag_name("a").get_attribute("href").split("/")
            )
            print("Found commenter: " + user_commented[3])
            user_commented_list.append(user_commented[3])

    except Exception as e:
        print("cant get comments" + str(e))

    date_time = browser.find_element_by_tag_name("time").get_attribute("datetime")
    return user_commented_list, date_time


def extract_information(browser, username, daysold, max_pic):
    """Get all the information for the given username"""
    web_address_navigator(browser, "https://www.instagram.com/" + username)

    try:
        num_of_posts = get_number_of_posts(browser)
        num_of_posts = min(num_of_posts, max_pic)
        # we don't need to scroll more than is max number of posts we want
        # to extract

    except Exception as e:
        print("\nError: Couldn't get user profile. Moving on.." + str(e))
        return []

    # PROFILE SCROLLING AND HARVESTING LINKS
    try:
        body_elem = browser.find_element_by_tag_name("body")

        links = []
        links2 = []
        links3 = []
        # list links contains 30 links from the current view, as that is the
        # maximum Instagram is showing at one time
        # list links2 contains all the links collected so far without
        # duplicates, in mixed order
        # list links3 contains all the links collected so far with
        # duplicates in preserved order
        previouslen = -1

        # every 60 links we will open picture and check it's date not to
        # scroll endlessly in huge profiles such as natgeo
        opened_overlay = 42
        sleep(0.5)

        # cycle that scrolls down the feed and collects links and saving
        # them into links2
        while len(links2) < num_of_posts:
            prev_divs = browser.find_elements_by_tag_name("main")
            # harvesting current img links:
            links_elems = [div.find_elements_by_tag_name("a") for div in prev_divs]
            links = sum(
                [
                    [link_elem.get_attribute("href") for link_elem in elems]
                    for elems in links_elems
                ],
                [],
            )
            # saving links for later:
            for link in links:
                if "/p/" in link:
                    links2.append(link)
                    links3.append(link)
            links2 = list(set(links2))
            # if after previous scroll, size of links2 didnt increase,
            # we should finish else we continue
            if len(links2) == previouslen:
                print("Cannot scroll, quitting..")
                sleep(0.5)
                break

            else:
                print("Scrolling profile ", len(links2), "/", num_of_posts)

                # TRYING TO END SCROLLING IN TIME
                # check the date of the image once in a 60 to not scroll too
                # much
                # only do it if we have a lot to images to go
                if (num_of_posts - len(links2) > 60) and (len(links2) > opened_overlay):
                    opened_overlay += 60

                    print("clicking on one photo..")
                    try:
                        one_pic_elem = browser.find_element_by_xpath(
                            read_xpath(extract_information.__name__, "one_pic_elem")
                        )
                        click_element(browser, one_pic_elem)
                    except Exception:
                        print("Error: cant click on the photo..")

                    sleep(1.5)

                    # following 6 lines give like to opened picture, to use
                    # our time effectively and look less suspicious
                    try:
                        like_element = browser.find_elements_by_xpath(
                            read_xpath(extract_information.__name__, "like_element")
                        )
                        click_element(browser, like_element[0])
                        print("clicking like..")
                    except Exception:
                        pass
                    sleep(2)

                    pic_date_time = browser.find_element_by_tag_name(
                        "time"
                    ).get_attribute("datetime")
                    pastdate = datetime.now() - timedelta(days=daysold)
                    date_of_pic = datetime.strptime(
                        pic_date_time, "%Y-%m-%dT%H:%M:%S.%fZ"
                    )

                    print("closing overlay")
                    close_overlay = browser.find_element_by_xpath(
                        read_xpath(extract_information.__name__, "close_overlay")
                    )
                    click_element(browser, close_overlay)

                    print("date of this picture was:", date_of_pic)

                    if date_of_pic < pastdate:
                        print("\nFinished scrolling, too old photos")
                        sleep(3)
                        break
                    else:
                        print("\nPhotos seems to be fresh, continuing scrolling")
                        sleep(2)

                previouslen = len(links2)
                body_elem = browser.find_element_by_tag_name("body")
                body_elem.send_keys(Keys.END)
                sleep(1.5)

    except (NoSuchElementException, StaleElementReferenceException) as err:
        print(
            "\n- Something went terribly wrong\n - Stopping everything and "
            "moving on with what I have\n"
        )
        print(err)

    links4 = remove_duplicates_preserving_order(links3)

    # PICTURES SCRAPPER ONE BY ONE
    # into user_commented_total_list go all username links who commented on
    # any post of this user
    counter = 1
    user_commented_total_list = []
    for link in links4:
        if max_pic <= 0:
            break
        max_pic -= 1
        print("\n", counter, " of max ", len(links4), " --- ", max_pic, " to go.")
        counter = counter + 1
        print("\nScrapping link: ", link)

        try:
            web_address_navigator(browser, link)
            user_commented_list, pic_date_time = extract_post_info(browser)
            user_commented_total_list = user_commented_total_list + user_commented_list

            # stop if date older than daysago
            pastdate = datetime.now() - timedelta(days=daysold)
            date_of_pic = datetime.strptime(pic_date_time, "%Y-%m-%dT%H:%M:%S.%fZ")
            print("date of pic: ", date_of_pic)
            if date_of_pic > pastdate:
                print("Recent pic, continue..")
            else:
                print("Old pic, ending getting users who commented.")
                sleep(3)
                break
            sleep(1)
        except NoSuchElementException as e:
            print("- Could not get information from post: " + link + " " + str(e))

    # PREPARE THE USER LIST TO EXPORT
    # sorts the list by frequencies, so users who comment the most are at
    # the top
    counter = collections.Counter(user_commented_total_list)
    com = sorted(counter.most_common(), key=itemgetter(1, 0), reverse=True)
    com = map(lambda x: [x[0]] * x[1], com)
    user_commented_total_list = [item for sublist in com for item in sublist]

    # remove duplicates preserving order (that's why not using set())
    user_commented_list = []
    last = ""
    for index, _ in enumerate(user_commented_total_list):
        if username.lower() != user_commented_total_list[index]:
            if (
                last != user_commented_total_list[index]
                and "p" not in user_commented_total_list[index]
            ):
                user_commented_list.append(user_commented_total_list[index])
            last = user_commented_total_list[index]

    print("\nGetting list of users who commented on this profile finished: ")
    print(user_commented_list, "\n")
    return user_commented_list


def users_liked(browser, photo_url, amount=100):
    photo_likers = []
    try:
        web_address_navigator(browser, photo_url)
        photo_likers = likers_from_photo(browser, amount)
        sleep(2)
    except NoSuchElementException:
        print("Could not get information from post: " + photo_url, " nothing to return")

    return photo_likers


def likers_from_photo(browser, amount=20):
    """ Get the list of users from the 'Likes' dialog of a photo """

    try:
        if check_exists_by_xpath(
            browser, read_xpath(likers_from_photo.__name__, "second_counter_button")
        ):
            liked_this = browser.find_elements_by_xpath(
                read_xpath(likers_from_photo.__name__, "second_counter_button")
            )
            element_to_click = liked_this[0]
        elif check_exists_by_xpath(
            browser, read_xpath(likers_from_photo.__name__, "liked_counter_button")
        ):

            liked_this = browser.find_elements_by_xpath(
                read_xpath(likers_from_photo.__name__, "liked_counter_button")
            )
            likers = []

            for liker in liked_this:
                if " like this" not in liker.text:
                    likers.append(liker.text)

            if " others" in liked_this[-1].text:
                element_to_click = liked_this[-1]

            elif " likes" in liked_this[0].text:
                element_to_click = liked_this[0]

            else:
                print(
                    "Few likes, not guaranteed you don't follow these"
                    " likers already.\nGot photo likers: {}\n".format(likers)
                )
                return likers

        else:
            print("Couldn't find liked counter button. May be a video.")
            print("Moving on..")
            return []

        sleep(1)
        click_element(browser, element_to_click)
        print("opening likes")
        # update server calls
        update_activity(browser, state=None)

        sleep(1)

        # get a reference to the 'Likes' dialog box
        dialog = browser.find_element_by_xpath(
            read_xpath("class_selectors", "likes_dialog_body_xpath")
        )

        # scroll down the page
        previous_len = -1
        browser.execute_script(
            "arguments[0].scrollTop = arguments[0].scrollHeight", dialog
        )
        update_activity(browser, state=None)
        sleep(1)

        start_time = time.time()
        user_list = []

        while (
            not user_list
            or (len(user_list) != previous_len)
            and (len(user_list) < amount)
        ):

            if previous_len + 10 >= amount:
                print("\nScrolling finished")
                if amount < 10:
                    user_list = get_users_from_dialog(user_list, dialog)
                sleep(1)
                break

            previous_len = len(user_list)
            scroll_bottom(browser, dialog, 2)

            user_list = get_users_from_dialog(user_list, dialog)

            # write & update records at Progress Tracker
            progress_tracker(len(user_list), amount, start_time, None)

        print("\n")
        random.shuffle(user_list)
        sleep(1)

        close_dialog_box(browser)

        print(
            "Got {} likers shuffled randomly whom you can follow:\n{}"
            "\n".format(len(user_list), user_list)
        )
        return user_list

    except Exception as exc:
        print("Some problem occured!\n\t{}".format(str(exc).encode("utf-8")))
        return []


def get_photo_urls_from_profile(
    browser, username, links_to_return_amount=1, randomize=True
):
    # try:
    # input can be both username or user profile url
    username = username_url_to_username(username)
    print("\nGetting likers from user: ", username, "\n")
    web_address_navigator(browser, "https://www.instagram.com/" + username + "/")
    sleep(1)

    photos_a_elems = browser.find_elements_by_xpath(
        read_xpath(get_photo_urls_from_profile.__name__, "photos_a_elems")
    )

    links = []
    for photo_element in photos_a_elems:
        photo_url = photo_element.get_attribute("href")
        # print ("photo url: ", photo_url)
        if "/p/" in photo_url:
            links.append(photo_url)

    if randomize is True:
        print("shuffling links")
        random.shuffle(links)
    print(
        "Got ",
        len(links),
        ", returning ",
        min(links_to_return_amount, len(links)),
        " links: ",
        links[:links_to_return_amount],
    )
    sleep(1)
    return links[:links_to_return_amount]
